/*----------------------------------------------------------------------------
 L3WPRINT - Harbour Minigui Win32 PRINT library source code

 Copyright 2010-2017 Luigi Ferraris <l3w-info@l3w.it>
 http://www.l3w.it/

 This program is free software; you can redistribute it and/or modify it under 
 the terms of the GNU General Public License as published by the Free Software 
 Foundation; either version 2 of the License, or (at your option) any later 
 version. 

 This program is distributed in the hope that it will be useful, but WITHOUT 
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

 You should have received a copy of the GNU General Public License along with 
 this software; see the file COPYING. If not, write to the Free Software 
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA (or 
 visit the web site http://www.gnu.org/).

 As a special exception, you have permission for additional uses of the text 
 contained in this release of L3WPRINT.

 The exception is that, if you link the L3WPRINT library with other
 files to produce an executable, this does not by itself cause the resulting 
 executable to be covered by the GNU General Public License.
 Your use of that executable is in no way restricted on account of linking the 
 L3WPRINT library code into it.

 Parts of this project are based upon:

	"MINIGUI - Harbour Win32 GUI Library"
        Copyright 2002-2009 Roberto Lopez <harbourminigui@gmail.com>
        http://harbourminigui.googlepages.com/

	"Harbour GUI framework for Win32"
 	Copyright 2001 Alexander S.Kresin <alex@belacy.belgorod.su>
 	Copyright 2001 Antonio Linares <alinares@fivetech.com>
	www - http://www.harbour-project.org

	"Harbour Project"
	Copyright 1999-2008, http://www.harbour-project.org/

	"WHAT32"
	Copyright 2002 AJ Wos <andrwos@aust1.net> 

	"HWGUI"
  	Copyright 2001-2008 Alexander S.Kresin <alex@belacy.belgorod.su>

---------------------------------------------------------------------------*/
#include "hmg.ch"
#include "hbclass.ch"


********************************************************************
* funzioni libreria
********************************************************************
CREATE CLASS L3WPrint
 VAR      lSelect      INIT .F.
 VAR      cError       INIT ""
 VAR      lError       INIT .F.

 VAR      cPrname      INIT ""
 VAR      cDocname     INIT "Untitled"
 VAR      nCopies      INIT 1
 VAR      nPagenum     INIT 0
 VAR      lPreview     INIT .T.
 VAR      nQuality     INIT PRINTER_RES_DRAFT
 VAR      nColor       INIT PRINTER_COLOR_MONOCHROME
 VAR      nOrientation INIT PRINTER_ORIENT_PORTRAIT
 VAR      nFrontrear   INIT PRINTER_DUP_SIMPLEX
 VAR      aPaper       INIT {} // HIDDEN
 VAR      nPaperSize   INIT PRINTER_PAPER_A4
 VAR      nPaperWidth  INIT 210 // VISIBLE
 VAR      nPaperHeight INIT 297 // VISIBLE
 VAR      nCollate     INIT PRINTER_COLLATE_FALSE

 VAR      lDocStarted  INIT .F.
 VAR      lDocClosed   INIT .F.
 VAR      lPageStarted INIT .F.
 VAR      lNewPage     INIT .F. // To prevent use of skip if a new page is started
 VAR      nCurRow      INIT 0
 VAR      cFontName    INIT "Courier New"
 VAR      nFontSize    INIT 10
 VAR      aMargin      INIT {0, 0, 0, 0}
 VAR      lByPass      INIT .F.    // to bypass EndOfPage Test for Headers and Footers
 VAR      aBdHead      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height
 VAR      aBdFoot      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height
 VAR      aPgHead      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height
 VAR      aPgFoot      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height
 VAR      lRpHeader    INIT .T.    // Report Header Must Be Print Only One Time: after printed out is set to .F.
 VAR      aRpHead      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height
 VAR      lRpFooter    INIT .F.    // Report Header Must Be Print Only One Time: when Enddoc Method is call is set to .T.
 VAR      aRpFoot      INIT { NIL, 0, 0 }    // Block, inirow[0|<>0], Height

 VAR      nRowHeight   INIT 5

 METHOD   New()
 METHOD   Select()
 METHOD   Setpro()
 METHOD   Getpro()
 METHOD   StartDoc()
 METHOD   EndDoc()
 METHOD   StartPage()
 METHOD   EndPage()

 METHOD   EndOfPage()
 METHOD   SkipLine()
 METHOD   GotoLine()
 METHOD   PrintText()
 METHOD   PrintRectangle()
 METHOD   PrintLine()
 METHOD   PrintImage()

 METHOD   Error()              INLINE ::lError
 METHOD   ErrorMsg()           INLINE ::cError

 METHOD   BodyHeader()         HIDDEN
 METHOD   BodyFooter()         HIDDEN
 METHOD   PageHeader()         HIDDEN
 METHOD   PageFooter()         HIDDEN
 METHOD   ReportHeader()       HIDDEN
 METHOD   ReportFooter()       HIDDEN
 METHOD   FooterHeight()       HIDDEN
 METHOD   FooterLimit()        HIDDEN

 METHOD   AllPrinters()        HIDDEN
 METHOD   Rotate()             HIDDEN
 METHOD   AddMrgToRow()        HIDDEN
 METHOD   SubMrgToRow()        HIDDEN
 METHOD   AddMrgToCol()        HIDDEN
 METHOD   SubMrgToCol()        HIDDEN

ENDCLASS

/*****************************************************************
* Create Class
*****************************************************************/
METHOD New() CLASS L3WPrint
LOCAL nI := 0
FOR nI:=1 TO 18
 AADD(::aPaper, {0,0} )
NEXT nI
::aPaper[ PRINTER_PAPER_LETTER 	    ] := { 215.9	, 279.4 }
::aPaper[ PRINTER_PAPER_LETTERSMALL ] := { 215.9	, 279.4 }
::aPaper[ PRINTER_PAPER_TABLOID     ] := { 279.4	, 431.8 }
::aPaper[ PRINTER_PAPER_LEDGER      ] := { 431.8	, 279.4 }
::aPaper[ PRINTER_PAPER_LEGAL       ] := { 215.9	, 355.6 }
::aPaper[ PRINTER_PAPER_STATEMENT   ] := { 139.7	, 215.9 }
::aPaper[ PRINTER_PAPER_EXECUTIVE   ] := { 184.15	, 266.7 }
::aPaper[ PRINTER_PAPER_A3          ] := { 297	, 420	}
::aPaper[ PRINTER_PAPER_A4          ] := { 210	, 297	}
::aPaper[ PRINTER_PAPER_A4SMALL     ] := { 210	, 297	}
::aPaper[ PRINTER_PAPER_A5          ] := { 148	, 210	}
::aPaper[ PRINTER_PAPER_B4          ] := { 250	, 354	}
::aPaper[ PRINTER_PAPER_B5          ] := { 182	, 257	}
::aPaper[ PRINTER_PAPER_FOLIO       ] := { 215.9	, 330.2	}
::aPaper[ PRINTER_PAPER_QUARTO      ] := { 215	, 275	}
::aPaper[ PRINTER_PAPER_10X14       ] := { 254	, 355.6	}
::aPaper[ PRINTER_PAPER_11X17       ] := { 279.4	, 431.8	}
::aPaper[ PRINTER_PAPER_NOTE        ] := { 215.9	, 279.4 }
::Rotate()

IF LEN(::AllPrinters())<1
 ::cError := "Non sono installate stampanti"
 ::lError := .T.
 ::lSelect := .F.
ENDIF

RETURN Self

/*****************************************************************
* Start Print: this Method is MANDATORY
*****************************************************************/
METHOD StartDoc() CLASS L3WPrint
IF ::lDocClosed == .F. .AND. ::lDocStarted == .F.
 IF ::lPreview
   SELECT PRINTER ::cPrname ;
   ORIENTATION ::nOrientation ;
   PAPERSIZE ::nPaperSize ;
   COPIES ::nCopies ;
   QUALITY ::nQuality ;
   COLOR ::nColor ;
   DUPLEX ::nFrontrear ;
   COLLATE ::nCollate ;
   PREVIEW
 ELSE
   SELECT PRINTER ::cPrname ;
   ORIENTATION ::nOrientation ;
   PAPERSIZE ::nPaperSize ;
   COPIES ::nCopies ;
   QUALITY ::nQuality ;
   COLOR ::nColor ;
   DUPLEX ::nFrontrear ;
   COLLATE ::nCollate
 ENDIF
 START PRINTDOC NAME ::cDocname
 ::lDocStarted := .T.
 ::lDocClosed := .F.
 ::lPageStarted := .F.
ENDIF
RETURN Self

/*****************************************************************
* End Print: this Method is MANDATORY
*****************************************************************/
METHOD EndDoc() CLASS L3WPrint
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T.
 IF ::lPageStarted == .T.
  ::lRpFooter := .T.
  ::Endpage()
 ENDIF
 END PRINTDOC
ENDIF
::lDocStarted := .F.
::lDocClosed := .T.
RETURN Self

/*****************************************************************
* Start Page: this Method is MANDATORY to print out, but you can omitt in you source
*****************************************************************/
METHOD StartPage() CLASS L3WPrint
IF ::lDocClosed == .F.
 IF ::lDocStarted == .F.
  ::Startdoc()
 ENDIF
 IF ::lPageStarted == .F.
  ::lPageStarted := .T.
  START PRINTPAGE
  ::lNewPage := .T.
  ::nPageNum += 1
  ::nCurRow := 0
  IF ::lRpHeader
   ::ReportHeader()
  ENDIF
  ::PageHeader()
  ::BodyHeader()
 ENDIF
ENDIF
RETURN Self

/*****************************************************************
* End Page: this Method is MANDATORY to print out, but you can omitt in you source
*****************************************************************/
METHOD EndPage() CLASS L3WPrint
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T.
 IF ::lPageStarted == .T.
  ::BodyFooter()
  ::PageFooter()
  IF ::lRpFooter == .T.
   ::ReportFooter()
  ENDIF
 ENDIF
 ::lNewPage := .F.
 END PRINTPAGE
 ::lPageStarted := .F.
ENDIF
RETURN Self

/*****************************************************************
* Body Header: this Method is optional.
* If you want to print out something BEFORE data you must activate it
* Example: data layout
*****************************************************************/
METHOD BodyHeader() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aBdHead[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aBdHead[2]==0, ::nCurRow, ::aBdHead[2] )
 xCurRow := ::nCurRow
 EVAL( ::aBdHead[1] )
 ::nCurRow := xCurRow + ::aBdHead[3]
 ::lByPass := .F.
ENDIF
::nRowHeight := xRowHeight
RETURN Self

/*****************************************************************
* Body Footer: this Method is optional
* If you want to print out something AFTER data you must activate it
* Example: data groups or ending layout
*****************************************************************/
METHOD BodyFooter() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aBdFoot[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aBdFoot[2]==0, ::nCurRow, ::aBdFoot[2] )
 xCurRow := ::nCurRow
 EVAL( ::aBdFoot[1] )
 ::nCurRow := xCurRow + ::aBdFoot[3]
 ::lByPass := .F.
 ENDIF
::nRowHeight := xRowHeight
RETURN Self

/*****************************************************************
* Page Header: this Method is optional.
* If you want to print out something BEFORE [BodyHeader and] data you must activate it
* Example: logo, title, data layout
*****************************************************************/
METHOD PageHeader() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aPgHead[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aPgHead[2]==0, ::nCurRow, ::aPgHead[2] )
 xCurRow := ::nCurRow
 EVAL( ::aPgHead[1] )
 ::nCurRow := xCurRow + ::aPgHead[3]
 ::lByPass := .F.
ENDIF
::nRowHeight := xRowHeight
RETURN Self

/*****************************************************************
* Page Footer: this Method is optional.
* If you want print out something AFTER [BodyFooter and] data you must activate it
* Example: logo, title, page counter
*****************************************************************/
METHOD PageFooter() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aPgFoot[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aPgFoot[2]==0, ::nCurRow, ::aPgFoot[2] )
 xCurRow := ::nCurRow
 EVAL( ::aPgFoot[1] )
 ::nCurRow := xCurRow + ::aPgFoot[3]
 ::lByPass := .F.
ENDIF
::nRowHeight := xRowHeight
RETURN Self

/*****************************************************************
* Report Header: this Method is optional.
* If you want to print out something BEFORE [PageHeader and] [BodyHeader and] data you must activate it
* Example: logo, title, data layout
*****************************************************************/
METHOD ReportHeader() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aRpHead[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aRpHead[2]==0, ::nCurRow, ::aRpHead[2] )
 xCurRow := ::nCurRow
 EVAL( ::aRpHead[1] )
 ::nCurRow := xCurRow + ::aRpHead[3]
 ::lByPass := .F.
ENDIF
::nRowHeight := xRowHeight
::lRpHeader := .F.
RETURN Self

/*****************************************************************
* Report Footer: this Method is optional.
* If you want print out something AFTER [BodyFooter and] [BodyFooter and] data you must activate it
* Example: logo, title, page counter
*****************************************************************/
METHOD ReportFooter() CLASS L3WPrint
LOCAL xCurRow := ::nCurRow
LOCAL xRowHeight := ::nRowHeight
IF ::lDocClosed == .F. .AND. ::lDocStarted == .T. .AND. ::lPageStarted == .T. .AND. VALTYPE(::aRpFoot[1])=="B"
 ::lByPass := .T.
 ::nCurRow := IIF( ::aRpFoot[2]==0, ::nCurRow, ::aRpFoot[2] )
 xCurRow := ::nCurRow
 EVAL( ::aRpFoot[1] )
 ::nCurRow := xCurRow + ::aRpFoot[3]
 ::lByPass := .F.
ENDIF
::nRowHeight := xRowHeight
RETURN Self

/*****************************************************************
* Return Footers Height
*****************************************************************/
METHOD FooterHeight() CLASS L3WPrint
LOCAL xHeight := 0
xHeight += IIF( VALTYPE(::aBdFoot[1])=="B", ::aBdFoot[3], 0 )
xHeight += IIF( VALTYPE(::aPgFoot[1])=="B", ::aPgFoot[3], 0 )
xHeight += IIF( VALTYPE(::aRpFoot[1])=="B", ::aRpFoot[3], 0 )
RETURN xHeight

/*****************************************************************
* Return Minimal Footer Limit
*****************************************************************/
METHOD FooterLimit() CLASS L3WPrint
LOCAL xLimit := ::nPaperHeight - ::aMargin[3]
* Report footer
IF VALTYPE(::aRpFoot[1])=="B" .AND. ::lRpFooter == .T.
 IF ::aRpFoot[2]==0
  xLimit := xLimit - ::aRpFoot[3]
 ELSEIF ::aRpFoot[2] < xLimit
  xLimit := ::aRpFoot[2]
 ENDIF
ENDIF
* Page footer
IF VALTYPE(::aPgFoot[1])=="B"
 IF ::aPgFoot[2]==0
  xLimit := xLimit - ::aPgFoot[3]
 ELSEIF ::aPgFoot[2] < xLimit
  xLimit := ::aPgFoot[2]
 ENDIF
ENDIF
* Body footer
IF VALTYPE(::aBdFoot[1])=="B"
 IF ::aBdFoot[2]==0
  xLimit := xLimit - ::aBdFoot[3]
 ELSEIF ::aBdFoot[2] < xLimit
  xLimit := ::aBdFoot[2]
 ENDIF
ENDIF
RETURN xLimit

/*****************************************************************
* End Of Page Reached ?
*****************************************************************/
METHOD EndOfPage(nHeight) CLASS L3WPrint
LOCAL lReturn := .F.
LOCAL xHeight := IIF( VALTYPE(nHeight)=="N", nHeight, ::nFontSize )
LOCAL xFooterHeight := ::FooterHeight()
LOCAL xFooterLimit := ::FooterLimit()
IF ::lByPass==.F.
* IF ::nCurRow + xHeight > xFooterLimit
 IF ::nCurRow + ::nRowHeight > xFooterLimit
  lReturn := .T.
 ENDIF
ENDIF
RETURN lReturn

/*****************************************************************
* Add height to the current Row position
*****************************************************************/
METHOD SkipLine(nHeight) CLASS L3WPrint
LOCAL xHeight := IIF( VALTYPE(nHeight)=="N", nHeight, ::nFontSize )
IF ::lNewPage
 ::lNewPage := .F.
ELSE
 ::nCurRow += xHeight
ENDIF
RETURN Self

/*****************************************************************
* Move cursor to specified Row
*****************************************************************/
METHOD GotoLine(nRow) CLASS L3WPrint
LOCAL xRow := IIF( VALTYPE(nRow)=="N", nRow, ::nCurRow )
::nCurRow := xRow
RETURN Self

/*****************************************************************
* Print Out Text
* If lAbsolute is TRUE paint at nRow value ELSE ::nCurRow (in this last case nRow value can be used like an -little- increment)
*****************************************************************/
METHOD PrintText( lAbsolute, cValue, nRow, nCol, nWidth, nHeight, cFontName, nFontSize, lFontBold, lFontItal, lFontUnde, lFontStri, aFontColo, cAlignmnt )
LOCAL xAbsolute := IIF( VALTYPE(lAbsolute) == "L", lAbsolute, .F. )
LOCAL xValue := IIF( VALTYPE(cValue) == "C", cValue, "" )
LOCAL xFontName := IIF( VALTYPE(cFontName)=="C" .AND. !EMPTY(cFontName), cFontName, ::cFontName )
LOCAL xFontSize := IIF( VALTYPE(nFontSize)=="N" .AND. nFontSize>0, nFontSize, ::nFontSize )
LOCAL xFontBold := IIF( VALTYPE(lFontBold)=="L", lFontBold, .F. )
LOCAL xFontItal := IIF( VALTYPE(lFontItal)=="L", lFontItal, .F. )
LOCAL xFontUnde := IIF( VALTYPE(lFontUnde)=="L", lFontUnde, .F. )
LOCAL xFontStri := IIF( VALTYPE(lFontStri)=="L", lFontStri, .F. )
LOCAL xFontColo := IIF( VALTYPE(aFontColo)=="A" .AND. LEN(aFontColo)==3, aFontColo, {0, 0, 0} )
LOCAL xAlignmnt := IIF( VALTYPE(cAlignmnt)=="C", IIF( UPPER(cAlignmnt)=="CENTER", "CENTER", IIF( UPPER(cAlignmnt)=="RIGHT", "RIGHT", "" ) ), "" )
LOCAL xWidth := IIF( VALTYPE(nWidth)=="N", nWidth, LEN(xValue) )
LOCAL xHeight := IIF( VALTYPE(nHeight)=="N", nHeight, xFontSize )
LOCAL xRow := IIF( VALTYPE(nRow)=="N", nRow, 0 )
LOCAL xCol := IIF( VALTYPE(nCol)=="N", nCol, 0 )
::StartPage()
IF ::EndOfPage(xHeight)
 ::EndPage()
 ::StartPage()
ENDIF
xRow := IIF( xAbsolute, ::AddMrgToRow(xRow), ::AddMrgToRow(::nCurRow+xRow) )
xCol := ::AddMrgToCol(xCol)
_HMG_PRINTER_H_MULTILINE_PRINT ( _HMG_SYSDATA [ 374 ], xRow, xCol, xRow+xHeight, xCol+xWidth, xFontName, xFontSize, xFontColo[1], xFontColo[2], xFontColo[3], xValue, xFontBold, xFontItal, xFontUnde, xFontStri, .T., .T., .T., xAlignmnt )
::lNewPage := .F.
RETURN Self

/*****************************************************************
* Print Out Rectangle
* IF lAbsolute is TRUE paint FROM nFromRow TO nToRow
* ELSE paint FROM ::nCurRow TO ::nCurRow + nToRow (nToRow is like a Height value)
*****************************************************************/
METHOD PrintRectangle( lAbsolute, nFromRow, nFromCol, nToRow, nToCol, nPenWidth, aPenColor )
LOCAL xAbsolute := IIF( VALTYPE(lAbsolute) == "L", lAbsolute, .F. )
LOCAL xPenWidth := IIF( VALTYPE(nPenWidth)=="N" .AND. nPenWidth>0, nPenWidth, 0.1 )
LOCAL xPenColor := IIF( VALTYPE(aPenColor)=="A" .AND. LEN(aPenColor)==3, aPenColor, {0, 0, 0} )
LOCAL xFromRow := IIF( VALTYPE(nFromRow)=="N", nFromRow, 0 )
LOCAL xFromCol := IIF( VALTYPE(nFromCol)=="N", nFromCol, 0 )
LOCAL xToRow := IIF( VALTYPE(nToRow)=="N", nToRow, 0 )
LOCAL xToCol := IIF( VALTYPE(nToCol)=="N", nToCol, 0 )
LOCAL xHeight := IIF( xAbsolute, xToRow-xFromRow, xToRow)
::StartPage()
IF ::EndOfPage(xHeight)
 ::EndPage()
 ::StartPage()
ENDIF
xFromRow := IIF( xAbsolute, ::AddMrgToRow(xFromRow), ::AddMrgToRow(::nCurRow) )
xToRow := IIF( xAbsolute, ::AddMrgToRow(xToRow), ::AddMrgToRow(::nCurRow) + xHeight )
_HMG_PRINTER_H_RECTANGLE( _HMG_SYSDATA [ 374 ] , xFromRow, xFromCol, xToRow, xToCol, xPenWidth, xPenColor[1], xPenColor[2], xPenColor[3], .T., .T. )
::lNewPage := .F.
RETURN Self

/*****************************************************************
* Print Out Line
* IF lAbsolute is TRUE paint FROM nFromRow TO nToRow
* ELSE paint FROM ::nCurRow TO ::nCurRow + nToRow (nToRow is like a Height value)
*****************************************************************/
METHOD PrintLine( lAbsolute, nFromRow, nFromCol, nToRow, nToCol, nPenWidth, aPenColor )
LOCAL xAbsolute := IIF( VALTYPE(lAbsolute) == "L", lAbsolute, .F. )
LOCAL xPenWidth := IIF( VALTYPE(nPenWidth)=="N" .AND. nPenWidth>0, nPenWidth, 0.1 )
LOCAL xPenColor := IIF( VALTYPE(aPenColor)=="A" .AND. LEN(aPenColor)==3, aPenColor, {0, 0, 0} )
LOCAL xFromRow := IIF( VALTYPE(nFromRow)=="N", nFromRow, 0 )
LOCAL xFromCol := IIF( VALTYPE(nFromCol)=="N", nFromCol, 0 )
LOCAL xToRow := IIF( VALTYPE(nToRow)=="N", nToRow, 0 )
LOCAL xToCol := IIF( VALTYPE(nToCol)=="N", nToCol, 0 )
LOCAL xHeight := IIF( xAbsolute, xToRow-xFromRow, xToRow)
::StartPage()
IF ::EndOfPage(xHeight)
 ::EndPage()
 ::StartPage()
ENDIF
xFromRow := IIF( xAbsolute, ::AddMrgToRow(xFromRow), ::AddMrgToRow(::nCurRow) )
xToRow := IIF( xAbsolute, ::AddMrgToRow(xToRow), ::AddMrgToRow(::nCurRow) + xHeight )
_HMG_PRINTER_H_LINE ( _HMG_SYSDATA [ 374 ] , xFromRow, xFromCol, xToRow, xToCol, xPenWidth, xPenColor[1], xPenColor[2], xPenColor[3], .T., .T. )
::lNewPage := .F.
RETURN Self

/*****************************************************************
* Print Out Image
* IF lAbsolute is TRUE paint AT nRow
* ELSE paint AT ::nCurRow (in this last case nRow value can be used like an -little- increment)
*****************************************************************/
METHOD PrintImage( lAbsolute, cValue, nRow, nCol, nHeight, nWidth, lStretch )
LOCAL xAbsolute := IIF( VALTYPE(lAbsolute) == "L", lAbsolute, .F. )
LOCAL xValue := cValue
LOCAL xWidth := IIF( VALTYPE(nWidth)=="N", nWidth, 0 )
LOCAL xHeight := IIF( VALTYPE(nHeight)=="N", nHeight, 0 )
LOCAL xRow := IIF( VALTYPE(nRow)=="N", nRow, 0 )
LOCAL xCol := IIF( VALTYPE(nCol)=="N", nCol, 0 )
LOCAL xStretch := IIF( VALTYPE(lStretch) == "L", lStretch, .F. )
::StartPage()
IF ::EndOfPage(xHeight)
 ::EndPage()
 ::StartPage()
ENDIF
xRow := IIF( xAbsolute, ::AddMrgToRow(xRow), ::AddMrgToRow(::nCurRow+xRow) )
xCol := ::AddMrgToCol(xCol)
_HMG_PRINTER_H_IMAGE ( _HMG_SYSDATA [ 374 ] , xValue , xRow, xCol, xHeight, xWidth, .T. )
::lNewPage := .F.
RETURN Self

/*****************************************************************
* Select a Printer
*****************************************************************/
METHOD Select() CLASS L3WPrint
LOCAL aPrinters := ::AllPrinters()
LOCAL aQualdes := {"DRAFT","LOW","MEDIUM","HIGH"}
LOCAL aColordes := {"MONO","COLOR"}
LOCAL aDuplex := {"SIMPLEX","VERTICAL","HORIZONTAL"}
LOCAL aCollate := {"Yes","No"}
SET NAVIGATION EXTENDED
DEFINE WINDOW l3wPform ;
 AT 0,0 ;
 WIDTH 500 ;
 HEIGHT 350 ;
 TITLE 'Printer selection' ;
 MODAL ;
 FONT "Arial" ;
 SIZE 10

* frame 1
 DEFINE FRAME Frame_1
  PARENT l3wPform
  ROW 10
  COL 10
  WIDTH 450
  HEIGHT 80
  CAPTION "Printers"
 END FRAME
* printer
 DEFINE COMBOBOX nPrinter
  PARENT l3wPform
  ROW 40
  COL 30
  WIDTH 400
  TOOLTIP 'Printer name'
  ONCHANGE  ::Setpro( "printer", aPrinters[l3wPform.nPrinter.Value])
  ITEMS aPrinters
  VALUE 1
  SORT .F.
 END COMBOBOX

* frame 2
 DEFINE FRAME Frame_2
  PARENT l3wPform
  ROW 100
  COL 10
  WIDTH 450
  HEIGHT 150
  CAPTION "Options"
 END FRAME
* Preview
 DEFINE CHECKBOX lPreview
  PARENT l3wPform
  ROW 130
  COL 30
  CAPTION "Preview"
  VALUE .T.
  WIDTH 100
  HEIGHT 25
  TOOLTIP "Preview"
  ONCHANGE ::Setpro( "preview", l3wPform.lPreview.Value)
 END CHECKBOX
* Copies
 DEFINE LABEL lbcopies
  PARENT l3wPform
  ROW 130
  COL 220
  VALUE "Copies"
  WIDTH 80
  HEIGHT 25
  ALIGNMENT Center
 END LABEL
 DEFINE SPINNER nCopies
  PARENT l3wPform
  ROW 130
  COL 305
  RANGEMAX 999
  RANGEMIN 1
  VALUE 1
  WIDTH 50
  HEIGHT 25
  TOOLTIP "Number of copies"
  ONCHANGE ::Setpro( "copies", l3wPform.nCopies.Value)
  WRAP .T.
  INCREMENT 1
 END SPINNER
* Quality
 DEFINE LABEL lbquality
  PARENT l3wPform
  ROW 160
  COL 30
  VALUE "Quality"
  WIDTH 80
  HEIGHT 25
  ALIGNMENT Center
 END LABEL
 DEFINE COMBOBOX nQuality
  PARENT l3wPform
  ROW 160
  COL 115
  WIDTH 100
  TOOLTIP 'Quality'
  ONCHANGE ::Setpro( "quality", aQualdes[l3wPform.nQuality.value])
  ITEMS aQualdes
  VALUE 1
  SORT .F.
 END COMBOBOX
* Color
 DEFINE LABEL lbcolor
  PARENT l3wPform
  ROW 160
  COL 220
  VALUE "Color"
  WIDTH 80
  HEIGHT 25
  ALIGNMENT Center
 END LABEL
 DEFINE COMBOBOX nColor
  PARENT l3wPform
  ROW 160
  COL 305
  WIDTH 100
  TOOLTIP 'Color'
  ONCHANGE ::Setpro( "color", aColordes[l3wPform.nColor.value])
  ITEMS aColordes
  VALUE 1
  SORT .F.
 END COMBOBOX
* FrontRear
 DEFINE LABEL lbrear
  PARENT l3wPform
  ROW 190
  COL 30
  VALUE "Front Rear"
  WIDTH 80
  HEIGHT 25
  ALIGNMENT Center
 END LABEL
 DEFINE COMBOBOX nFront
  PARENT l3wPform
  ROW 190
  COL 115
  WIDTH 120
  TOOLTIP 'Front Rear'
  ONCHANGE ::Setpro( "frontrear", aDuplex[l3wPform.nFront.value])
  ITEMS aDuplex
  VALUE 1
  SORT .F.
 END COMBOBOX
* collate
 DEFINE CHECKBOX lCollate
  PARENT l3wPform
  ROW 190
  COL 305
  CAPTION "Collate"
  VALUE .T.
  WIDTH 100
  HEIGHT 25
  TOOLTIP "Collate"
  ONCHANGE ::Setpro( "collate", IIF(l3wPform.lCollate.Value==.T., 1, 0) )
 END CHECKBOX

* buttons
 DEFINE BUTTON bt_print
  PARENT l3wPform
  ROW 260
  COL 100
  PICTURE "Images/Print.bmp"
  ONCLICK {|| ::lSelect:=.T., l3wPform.Release }
  WIDTH 100
  HEIGHT 40
  TOOLTIP "Print"
  FLAT .T.
 END BUTTON
 DEFINE BUTTON bt_exit
  PARENT l3wPform
  ROW 260
  COL 300
  PICTURE "Images/Exit.bmp"
  ONCLICK {|| ::lSelect:=.F., l3wPform.Release }
  WIDTH 100
  HEIGHT 40
  TOOLTIP "Exit"
  FLAT .T.
 END BUTTON

END WINDOW
* Preset values
l3wPform.nPrinter.Value := IIF( ASCAN(aPrinters, GetDefaultPrinter())!=0, ASCAN(aPrinters, GetDefaultPrinter()), 1)
::Setpro( "printer", aPrinters[l3wPform.nPrinter.Value] )
l3wPform.lPreview.Value := ::lPreview
l3wPform.nCopies.Value := ::nCopies

IF ::nQuality==PRINTER_RES_DRAFT
 l3wPform.nQuality.value := 1
ELSEIF ::nQuality==PRINTER_RES_LOW
 l3wPform.nQuality.value := 2
ELSEIF ::nQuality==PRINTER_RES_MEDIUM
 l3wPform.nQuality.value := 3
ELSEIF ::nQuality==PRINTER_RES_HIGH
 l3wPform.nQuality.value := 4
ENDIF

IF ::nColor==PRINTER_COLOR_MONOCHROME
 l3wPform.nColor.value := 1
ELSEIF ::nColor==PRINTER_COLOR_COLOR
 l3wPform.nColor.value := 2
ENDIF

IF ::nFrontrear == PRINTER_DUP_SIMPLEX
 l3wPform.nFront.value := 1
ELSEIF ::nFrontrear := PRINTER_DUP_VERTICAL
 l3wPform.nFront.value := 2
ELSEIF ::nFrontrear := PRINTER_DUP_HORIZONTAL
 l3wPform.nFront.value := 3
ENDIF

IF ::nCollate == PRINTER_COLLATE_FALSE
 l3wPform.lCollate.Value := .F.
ELSEIF ::nCollate == PRINTER_COLLATE_TRUE
 l3wPform.lCollate.Value := .T.
ENDIF

l3wPform.Center
l3wPform.Activate
RETURN ::lSelect

/*****************************************************************
* Setproperty
* Used to set parameters, but someones ONLY before StartDoc
*****************************************************************/
METHOD Setpro(Arg1,Arg2) CLASS L3WPrint
IF VALTYPE(Arg1)=="C"
 Arg1 := LOWER(ALLTRIM(Arg1))
* Printer name
 IF Arg1 == "printer" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   ::cPrname := Arg2
  ENDIF
 ENDIF
* Document name
 IF Arg1 == "docname" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   ::cDocName := Arg2
  ENDIF
 ENDIF
* Copies number
 IF Arg1 == "copies" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="N" .AND. Arg2>=1
   ::nCopies := Arg2
  ENDIF
 ENDIF
* Preview
 IF Arg1 == "preview" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="L"
   ::lPreview := Arg2
  ENDIF
 ENDIF
* Quality
 IF Arg1 == "quality" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   DO CASE
    CASE LOWER(ALLTRIM(Arg2))=="draft"
     ::nQuality := PRINTER_RES_DRAFT
    CASE LOWER(ALLTRIM(Arg2))=="low"
     ::nQuality := PRINTER_RES_LOW
    CASE LOWER(ALLTRIM(Arg2))=="medium"
     ::nQuality := PRINTER_RES_MEDIUM
    CASE LOWER(ALLTRIM(Arg2))=="high"
     ::nQuality := PRINTER_RES_HIGH
   ENDCASE
  ENDIF
  IF VALTYPE(Arg2)=="N"
   DO CASE
    CASE Arg2==PRINTER_RES_DRAFT
     ::nQuality := PRINTER_RES_DRAFT
    CASE Arg2==PRINTER_RES_LOW
     ::nQuality := PRINTER_RES_LOW
    CASE Arg2==PRINTER_RES_MEDIUM
     ::nQuality := PRINTER_RES_MEDIUM
    CASE Arg2==PRINTER_RES_HIGH
     ::nQuality := PRINTER_RES_HIGH
   ENDCASE
  ENDIF
 ENDIF
* Color
 IF Arg1 == "color" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   DO CASE
    CASE LOWER(ALLTRIM(Arg2))=="mono"
     ::nColor := PRINTER_COLOR_MONOCHROME
    CASE LOWER(ALLTRIM(Arg2))=="color"
     ::nColor := PRINTER_COLOR_COLOR
   ENDCASE
  ENDIF
  IF VALTYPE(Arg2)=="N"
   DO CASE
    CASE Arg2==PRINTER_COLOR_MONOCHROME
     ::nColor := PRINTER_COLOR_MONOCHROME
    CASE Arg2==PRINTER_COLOR_COLOR
     ::nColor := PRINTER_COLOR_COLOR
   ENDCASE
  ENDIF
 ENDIF
* Orientation
 IF Arg1 == "orientation" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   DO CASE
    CASE UPPER(ALLTRIM(Arg2))=="portrait"
     ::nOrientation := PRINTER_ORIENT_PORTRAIT
    CASE UPPER(ALLTRIM(Arg2))=="landscape"
     ::nOrientation := PRINTER_ORIENT_LANDSCAPE
   ENDCASE
  ENDIF
  IF VALTYPE(Arg2)=="N"
   DO CASE
    CASE Arg2==PRINTER_ORIENT_PORTRAIT
     ::nOrientation := PRINTER_ORIENT_PORTRAIT
    CASE Arg2==PRINTER_ORIENT_LANDSCAPE
     ::nOrientation := PRINTER_ORIENT_LANDSCAPE
   ENDCASE
   ::Rotate()
  ENDIF
 ENDIF
* Frontrear
 IF Arg1 == "frontrear" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   DO CASE
    CASE LOWER(ALLTRIM(Arg2))=="simplex"
     ::nFrontrear := PRINTER_DUP_SIMPLEX
    CASE UPPER(ALLTRIM(Arg2))=="vertical"
     ::nFrontrear := PRINTER_DUP_VERTICAL
    CASE UPPER(ALLTRIM(Arg2))=="horizontal"
     ::nFrontrear := PRINTER_DUP_HORIZONTAL
   ENDCASE
  ENDIF
  IF VALTYPE(Arg2)=="N"
   DO CASE
    CASE Arg2==PRINTER_DUP_SIMPLEX
     ::nFrontrear := PRINTER_DUP_SIMPLEX
    CASE Arg2==PRINTER_DUP_VERTICAL
     ::nFrontrear := PRINTER_DUP_VERTICAL
    CASE Arg2==PRINTER_DUP_VERTICAL
     ::nFrontrear := PRINTER_DUP_HORIZONTAL
   ENDCASE
  ENDIF
 ENDIF
* Paper size
 IF Arg1 == "papersize" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="N" .AND. Arg2>=1 .AND. Arg2<=LEN(::aPaper)
   ::nPaperSize := Arg2
   ::Rotate()
  ENDIF
 ENDIF
* Collate
 IF Arg1 == "collate" .AND. ::lDocStarted==.F.
 IF VALTYPE(Arg2)=="N"
  DO CASE
   CASE Arg2==PRINTER_COLLATE_FALSE
    ::nCollate := Arg2
   CASE Arg2==PRINTER_COLLATE_TRUE
    ::nCollate := Arg2
   ENDCASE
  ENDIF
 ENDIF
* Cur Row
 IF Arg1 == "currow"
  IF VALTYPE(Arg2)=="N" .AND. Arg2>=0
   ::nCurRow := Arg2
  ENDIF
 ENDIF
* Font Name
 IF Arg1 == "fontname" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="C" .AND. !EMPTY(Arg2)
   ::cFontName := Arg2
  ENDIF
 ENDIF
* Font Size
 IF Arg1 == "fontsize" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="N" .AND. Arg2>=0
   ::nFontSize := Arg2
  ENDIF
 ENDIF
* Maring (style Html: Top, Right, Bottom, Left)
 IF Arg1 == "margin" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==4
   ::aMargin := ACLONE(Arg2)
  ENDIF
 ENDIF
* Body Header
 IF Arg1 == "bodyheader" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aBdHead := ACLONE(Arg2)
  ENDIF
 ENDIF
* Body Footer
 IF Arg1 == "bodyfooter" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aBdFoot := ACLONE(Arg2)
  ENDIF
 ENDIF
* Page Header
 IF Arg1 == "pageheader" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aPgHead := ACLONE(Arg2)
  ENDIF
 ENDIF
* Page Footer
 IF Arg1 == "pagefooter" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aPgFoot := ACLONE(Arg2)
  ENDIF
 ENDIF
* Report Header
 IF Arg1 == "repoheader" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aRpHead := ACLONE(Arg2)
  ENDIF
 ENDIF
* Report Footer
 IF Arg1 == "repofooter" .AND. ::lDocStarted==.F.
  IF VALTYPE(Arg2)=="A" .AND. LEN(Arg2)==3 .AND. ;
     VALTYPE(Arg2[1])=="B" .AND. ;
     VALTYPE(Arg2[2])=="N" .AND. Arg2[2]>=0 .AND. ;
     VALTYPE(Arg2[3])=="N" .AND. Arg2[3]>0
   ::aRpFoot := ACLONE(Arg2)
  ENDIF
 ENDIF

* Detail Height
 IF Arg1 == "rowheight"
  IF VALTYPE(Arg2)=="N" .AND. Arg2>=0
   ::nRowHeight := Arg2
  ENDIF
 ENDIF
ENDIF
RETURN Self

/*****************************************************************
* Getproperty
* Used to retrieve settings
*****************************************************************/
METHOD Getpro(Arg1) CLASS L3WPrint
LOCAL xReturn
IF VALTYPE(Arg1)=="C"
 Arg1 := LOWER(ALLTRIM(Arg1))
 IF Arg1 == "printer"
  xReturn := ::cPrname
 ENDIF
* Copies number
 IF Arg1 == "copies"
  xReturn := ::nCopies
 ENDIF
* Page Number
 IF Arg1 == "pagenum"
  xReturn := ::nPagenum
 ENDIF
* Preview
 IF Arg1 == "preview"
  xReturn := ::lPreview
 ENDIF
* Quality
 IF Arg1 == "quality"
  xReturn := ::nQuality
 ENDIF
* Color
 IF Arg1 == "color"
  xReturn := ::nColor
 ENDIF
* Orientation
 IF Arg1 == "orientation"
  xReturn := ::nOrientation
 ENDIF
* Frontrear
 IF Arg1 == "frontrear"
  xReturn := ::nFrontrear
 ENDIF
* Paper size
 IF Arg1 == "papersize"
  xReturn := ::nPaperSize
 ENDIF
* Paper Width
 IF Arg1 == "paperwidth"
  xReturn := ::nPaperWidth
 ENDIF
* Paper Height
 IF Arg1 == "paperheight"
  xReturn := ::nPaperHeight
 ENDIF
* Collate
 IF Arg1 == "collate"
  xReturn := ::nCollate
 ENDIF
* Doc Started
 IF Arg1 == "docstarted"
  xReturn := ::lDocStarted
 ENDIF
* Doc Closed
 IF Arg1 == "docclosed"
  xReturn := ::lDocClosed
 ENDIF
* Page Started
 IF Arg1 == "pagestarted"
  xReturn := ::lPageStarted
 ENDIF
* Cur Row
 IF Arg1 == "currow"
  xReturn := ::nCurRow
 ENDIF
* Min Row
 IF Arg1 == "minrow"
  xReturn := ::aMargin[1]
 ENDIF
* Max Row
 IF Arg1 == "maxrow"
  xReturn := ::nPaperHeight - ::aMargin[3]
 ENDIF
* Min Col
 IF Arg1 == "mincol"
  xReturn := ::aMargin[4]
 ENDIF
* Max Col
 IF Arg1 == "maxcol"
  xReturn := ::nPaperWidth - ::aMargin[2]
 ENDIF
* Font Name
 IF Arg1 == "fontname"
  xReturn := ::cFontName
 ENDIF
* Font Size
 IF Arg1 == "fontsize"
  xReturn := ::cFontSize
 ENDIF
* Maring (style Html: Top, Right, Bottom, Left)
 IF Arg1 == "margin"
  xReturn := ::aMargin
 ENDIF
* Body Header
 IF Arg1 == "bodyheader"
  xReturn := ::aBdHead
 ENDIF
* Body Footer
 IF Arg1 == "Bodyfooter"
  xReturn := ::aBdFoot
 ENDIF
* Page Header
 IF Arg1 == "pageheader"
  xReturn := ::aPgHead
 ENDIF
* Page Footer
 IF Arg1 == "pagefooter"
  xReturn := ::aPgFoot
 ENDIF
* Report Header
 IF Arg1 == "repoheader"
  xReturn := ::aRpHead
 ENDIF
* Report Footer
 IF Arg1 == "repofooter"
  xReturn := ::aRpFoot
 ENDIF
* Detail Height
 IF Arg1 == "rowheight"
  xReturn := ::nRowHeight
 ENDIF
ENDIF
RETURN xReturn

/*****************************************************************
* Add Top Margin to Row
* If Row it's omitted it's assumed current
*****************************************************************/
METHOD AddMrgToRow(nRw) CLASS L3WPrint
LOCAL xRow := IIF(VALTYPE(nRw)=="N", nRw, ::nCurRow)
xRow += ::aMargin[1]
RETURN xRow

/*****************************************************************
* Subtract Bottom Margin from Row
* If Row it's omitted it's assumed current
*****************************************************************/
METHOD SubMrgToRow(nRw) CLASS L3WPrint
LOCAL xRow := IIF(VALTYPE(nRw)=="N", nRw, ::nCurRow)
xRow -= ::aMargin[3]
RETURN xRow

/*****************************************************************
* Add Left Margin to Col
* If Col it's omitted it's assumed 0
*****************************************************************/
METHOD AddMrgToCol(nCl) CLASS L3WPrint
LOCAL xCol := IIF(VALTYPE(nCl)=="N", nCl, 0)
xCol += ::aMargin[4]
RETURN xCol

/*****************************************************************
* Subtract Right Margin from Col
* If Col it's omitted it's assumed 0
*****************************************************************/
METHOD SubMrgToCol(nCl) CLASS L3WPrint
LOCAL xCol := IIF(VALTYPE(nCl)=="N", nCl, 0)
xCol -= ::aMargin[2]
RETURN xCol

/*****************************************************************
* When Orientation or PaperSize were canghed
* reset PaperWidth and PaperHeight dimensions
*****************************************************************/
METHOD Rotate() CLASS L3WPrint
IF ::nOrientation == PRINTER_ORIENT_PORTRAIT
 ::nPaperWidth  := ::aPaper[ ::nPaperSize ] [ 1 ]
 ::nPaperHeight := ::aPaper[ ::nPaperSize ] [ 2 ]
ELSEIF ::nOrientation == PRINTER_ORIENT_LANDSCAPE
 ::nPaperWidth  := ::aPaper [ ::nPaperSize ] [ 2 ]
 ::nPaperHeight := ::aPaper [ ::nPaperSize ] [ 1 ]
ENDIF
RETURN Self

/*****************************************************************
* Find installed Printers
*****************************************************************/
METHOD AllPrinters() CLASS L3WPrint
RETURN aPrinters()
